@@ -1,9 +1,9 @@
-/* Copyright 2000-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -30,8 +30,19 @@ const struct testlist {
     {testpass},
     {testmd4},
     {testmd5},
+    {testcrypto},
     {testldap},
-    {testdbd}
+    {testdbd},
+    {testdate},
+    {testmemcache},
+    {testredis},
+    {testxml},
+    {testxlate},
+    {testrmm},
+    {testdbm},
+    {testqueue},
+    {testreslist},
+    {testsiphash}
 };
 
 #endif /* APR_TEST_INCLUDES */
