@@ -258,10 +258,10 @@ static void test_memcache_meta(abts_case * tc, void *data)
     rv = apr_memcache_add_server(memcache, server);
     ABTS_ASSERT(tc, "server add failed", rv == APR_SUCCESS);
 
-    rv = apr_memcache_version(server, pool, &result);
+    apr_memcache_version(server, pool, &result);
     ABTS_PTR_NOTNULL(tc, result);
 
-    rv = apr_memcache_stats(server, p, &stats);
+    apr_memcache_stats(server, p, &stats);
     ABTS_PTR_NOTNULL(tc, stats);
 
     ABTS_STR_NEQUAL(tc, stats->version, result, 5);
@@ -450,7 +450,7 @@ static void test_memcache_multiget(abts_case * tc, void *data)
     ABTS_ASSERT(tc, "set failed", rv == APR_SUCCESS);
   }
   
-  rv = apr_pool_create(&tmppool, pool);
+  apr_pool_create(&tmppool, pool);
   for (i = 0; i < TDATA_SET; i++)
     apr_memcache_add_multget_key(pool,
                                  apr_pstrcat(pool, prefix,
@@ -604,10 +604,10 @@ abts_suite *testmemcache(abts_suite * suite)
     apr_status_t rv;
     suite = ADD_SUITE(suite);
     /* check for a running memcached on the typical port before 
-     * trying to run the tests. succeed silently if we don't find one.
+     * trying to run the tests. succeed if we don't find one.
      */
     rv = check_mc();
-    if(rv == APR_SUCCESS) {
+    if (rv == APR_SUCCESS) {
       abts_run_test(suite, test_memcache_create, NULL);
       abts_run_test(suite, test_memcache_user_funcs, NULL);
       abts_run_test(suite, test_memcache_meta, NULL);
@@ -616,6 +616,11 @@ abts_suite *testmemcache(abts_suite * suite)
       abts_run_test(suite, test_memcache_addreplace, NULL);
       abts_run_test(suite, test_memcache_incrdecr, NULL);
     }
+    else {
+        abts_log_message("Error %d occurred attempting to reach memcached "
+                         "on %s:%d.  Skipping apr_memcache tests...",
+                         rv, HOST, PORT);
+    }
 
     return suite;
 }
